/**
 *  Copyright (c) 2008-2011 Sonatype, Inc. All rights reserved.
 *
 *  This program is licensed to you under the Apache License Version 2.0,
 *  and you may not use this file except in compliance with the Apache License Version 2.0.
 *  You may obtain a copy of the Apache License Version 2.0 at http://www.apache.org/licenses/LICENSE-2.0.
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the Apache License Version 2.0 is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the Apache License Version 2.0 for the specific language governing permissions and limitations there under.
 */
package org.sonatype.jira;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Singleton;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.fileupload.FileUploadException;
import org.apache.commons.fileupload.disk.DiskFileItem;
import org.apache.commons.fileupload.disk.DiskFileItemFactory;
import org.apache.commons.fileupload.servlet.ServletFileUpload;
import org.apache.xmlrpc.XmlRpcException;
import org.apache.xmlrpc.server.XmlRpcHandlerMapping;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.sonatype.jira.internal.JIRAHandlerMapping;

import com.ning.http.util.Base64;

@Singleton
@Named
public class MediatorServlet
    extends org.apache.xmlrpc.webserver.XmlRpcServlet
{
    private static final long serialVersionUID = 1709019220162713367L;

    private static final Logger logger = LoggerFactory.getLogger( MediatorServlet.class );

    private final JiraXmlRpcProxy handler;

    private final URL forwardUrl;

    private String version;

    private final List<AttachmentHandler> attachmentHandlers;

    @Inject
    public MediatorServlet( final JiraXmlRpcProxy handler, @Named( "${jira.proxy.target.url}" ) final URL url,
                            final List<AttachmentHandler> attachmentHandlers )
    {
        this.forwardUrl = url;
        this.handler = handler;
        this.attachmentHandlers = attachmentHandlers;
        handler.setProxiedUrl( url.toString() + "/rpc/xmlrpc" );
    }

    @Override
    public void doPost( final HttpServletRequest pRequest, final HttpServletResponse pResponse )
        throws IOException, ServletException
    {
        try
        {
            String pathInfo = pRequest.getPathInfo();
            if ( "/rpc/xmlrpc".equals( pathInfo ) )
            {
                super.doPost( pRequest, pResponse );
            }
            else if ( pathInfo.startsWith( "/secure/Attach" ) )
            {
                String username = pRequest.getParameter( "os_username" );
                String password = pRequest.getParameter( "os_password" );
                String issueKey = pRequest.getParameter( "issue_key" );
                String issueId = pRequest.getParameter( "id" );

                postAttachments( pRequest, username, password, issueKey, issueId );
            }
            else if ( pathInfo.startsWith( "/rest/api/latest/issue/" ) && pathInfo.endsWith( "/attachments" ) )
            {
                String[] credentials = getCredentials( pRequest );

                String username = credentials[0];
                String password = credentials[1];

                String[] parts = pathInfo.split( "/" );
                if ( parts.length < 6 )
                {
                    throw new MediatorException( "No issue key found in REST path: " + pathInfo );
                }
                String issueKey = parts[5];

                postAttachments( pRequest, username, password, issueKey, null );
            }
            else
            {
                pResponse.sendRedirect( forwardUrl + pathInfo );
            }
        }
        catch ( Throwable t )
        {
            logger.debug( t.getMessage(), t );
            pResponse.sendError( 500, t.getMessage() );
        }
    }

    @Override
    protected void doGet( final HttpServletRequest req, final HttpServletResponse resp )
        throws ServletException, IOException
    {
        resp.sendRedirect( forwardUrl + req.getPathInfo() );
    }

    private String[] getCredentials( final HttpServletRequest pRequest )
        throws IOException
    {
        String authHeader = pRequest.getHeader( "Authorization" );
        String[] credentials;

        String username = pRequest.getParameter( "os_username" );
        String password = pRequest.getParameter( "os_password" );
        if ( username != null )
        {
            credentials = new String[] { username, password };
        }
        else if ( authHeader != null )
        {
            if ( !authHeader.startsWith( "Basic " ) )
            {
                throw new MediatorException( "Can only handle Basic authentication for REST resources." );
            }

            try
            {
                String decoded = new String( Base64.decode( authHeader.substring( "Basic ".length() ) ), "utf-8" );
                credentials = decoded.split( ":", 2 );
            }
            catch ( RuntimeException e )
            {
                throw new MediatorException( "Could not decode Basic authentication header", e );
            }
        }
        else
        {
            credentials = new String[] { null, null };
        }
        return credentials;
    }

    private void postAttachments( final HttpServletRequest pRequest, final String username, final String password,
                                  final String issueKey, final String issueId )
        throws IOException
    {
        String jiraVersion = getRemoteVersion( pRequest );
        Map<DiskFileItem, Attachment> attachments = getAttachments( pRequest );
        postAttachments( username, password, issueKey, issueId, jiraVersion, attachments );
    }

    private void postAttachments( final String username, final String password, final String issueKey,
                                  final String issueId, final String jiraVersion,
                                  final Map<DiskFileItem, Attachment> attachments )
        throws IOException
    {
        Set<DiskFileItem> items = null;
        try
        {
            items = attachments.keySet();

            AttachmentHandler handler = chooseAttachmentHandler( jiraVersion );

            AttachmentResponse attachmentResponse =
                handler.attach( forwardUrl.toString(), username, password, issueId, issueKey, attachments.values() );

            if ( attachmentResponse.isErrorResponse() )
            {
                throw new MediatorException( "Could not upload attachments: " + attachmentResponse.getStatusText() );
            }
        }
        finally
        {
            if ( items != null ) // exception while getting attachments
            {
                for ( DiskFileItem item : items )
                {
                    item.delete();
                }
            }
        }
    }

    @SuppressWarnings( "unchecked" )
    private Map<DiskFileItem, Attachment> getAttachments( final HttpServletRequest pRequest )
        throws IOException
    {
        Map<DiskFileItem, Attachment> attachments = new HashMap<DiskFileItem, Attachment>();

        ServletFileUpload upload =
            new ServletFileUpload( new DiskFileItemFactory( 0, new File( System.getProperty( "java.io.tmpdir" ) ) ) );

        List<DiskFileItem> items;
        try
        {
            items = upload.parseRequest( pRequest );
            for ( final DiskFileItem item : items )
            {
                attachments.put( item, new Attachment()
                {

                    @Override
                    public String getName()
                    {
                        return item.getName();
                    }

                    @Override
                    public long getLength()
                    {
                        return item.getSize();
                    }

                    @Override
                    public InputStream getInputStream()
                        throws IOException
                    {
                        return item.getInputStream();
                    }

                    @Override
                    public String getContentType()
                    {
                        return item.getContentType();
                    }
                } );
            }
        }
        catch ( FileUploadException e )
        {
            throw (IOException) new IOException( e.getMessage() ).initCause( e );
        }
        return attachments;
    }

    private AttachmentHandler chooseAttachmentHandler( final String jiraVersion )
    {
        for ( AttachmentHandler handler : attachmentHandlers )
        {
            if ( handler.isSupporting( jiraVersion ) )
            {
                return handler;
            }
        }

        throw new MediatorException( "Could not find attachment handler for JIRA version " + jiraVersion );
    }

    private String getRemoteVersion( final HttpServletRequest pRequest )
        throws IOException
    {
        if ( this.version == null )
        {
            String[] credentials = getCredentials( pRequest );

            String username = credentials[0];
            String password = credentials[1];

            String token = handler.login( username, password );

            @SuppressWarnings( "rawtypes" )
            Hashtable info = handler.getServerInfo( token );

            System.err.println( info );
            this.version = info.get( "version" ).toString();
        }

        return "JIRA " + this.version;
    }

    @Override
    protected XmlRpcHandlerMapping newXmlRpcHandlerMapping()
        throws XmlRpcException
    {
        return new JIRAHandlerMapping( handler );
    }

}