/**
 *  Copyright (c) 2008-2011 Sonatype, Inc. All rights reserved.
 *
 *  This program is licensed to you under the Apache License Version 2.0,
 *  and you may not use this file except in compliance with the Apache License Version 2.0.
 *  You may obtain a copy of the Apache License Version 2.0 at http://www.apache.org/licenses/LICENSE-2.0.
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the Apache License Version 2.0 is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the Apache License Version 2.0 for the specific language governing permissions and limitations there under.
 */
package org.sonatype.jira.internal;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.inject.Inject;
import javax.inject.Named;

import org.apache.xmlrpc.XmlRpcException;
import org.apache.xmlrpc.XmlRpcRequest;
import org.apache.xmlrpc.XmlRpcRequestConfig;
import org.apache.xmlrpc.client.XmlRpcAhcTransportFactory;
import org.apache.xmlrpc.client.XmlRpcClient;
import org.apache.xmlrpc.client.XmlRpcClientConfigImpl;
import org.apache.xmlrpc.client.XmlRpcClientRequestImpl;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.sonatype.inject.Nullable;
import org.sonatype.jira.JiraXmlRpc;
import org.sonatype.jira.JiraXmlRpcProxy;
import org.sonatype.jira.recorder.Recorder;
import org.sonatype.jira.recorder.RecordingProxy;
import org.sonatype.jira.recorder.Triple;

/**
 * Proxy for JIRA 4.1.2 XMLRPC calls, just delegates the call to a remote server.
 */
@Named
@SuppressWarnings( "rawtypes" )
public class DefaultJiraXmlRpcProxy
    implements InvocationHandler, JiraXmlRpcProxy, RecordingProxy
{

    private final Logger logger = LoggerFactory.getLogger( this.getClass() );

    private final JiraXmlRpc proxy;

    private XmlRpcClient client;

    private URL url;

    private final List<Triple> triplets = new LinkedList<Triple>();

    private boolean recording = false;

    public DefaultJiraXmlRpcProxy()
    {
        this.proxy = (JiraXmlRpc) Proxy.newProxyInstance( this.getClass().getClassLoader(), new Class[] { JiraXmlRpc.class }, this );
    }

    @Inject
    public DefaultJiraXmlRpcProxy( @Nullable @Named( "${jira.proxy.target.url}" ) final String url )
    {
        if ( url != null && !"${jira.proxy.target.url}".equals( url ) )
        {
            setProxiedUrl( url );
        }
        this.proxy = (JiraXmlRpc) Proxy.newProxyInstance( this.getClass().getClassLoader(), new Class[] { JiraXmlRpc.class }, this );
    }

    @Override
    public String getProxiedUrl()
    {
        return url == null ? "" : url.toString();
    }

    @Override
    public Object invoke( final Object proxy, final Method method, final Object[] args )
        throws Throwable
    {
        String methodName = method.getName();
        logger.trace( "calling rpc method: {}#{}({})", new Object[] { url, methodName, Arrays.toString( args ) } );

        XmlRpcClient c = client();
        XmlRpcRequest request = request( methodName, args );
        Object result = null;
        try
        {
            result = c.execute( request );
        }
        catch ( XmlRpcException e )
        {
            if ( recording )
            {
                triplets.add( new Triple( method, args, e ) );
            }
            throw e;
        }

        logger.trace( "received rpc result for method: {}#{}({})\n{}",
                      new Object[] { url, methodName, Arrays.toString( args ), result } );

        if ( result instanceof Object[] )
        {
            Vector<Object> returnValue = new Vector<Object>();
            Object[] resultArray = (Object[]) result;
            for ( Object o : resultArray )
            {
                if ( o instanceof Map )
                {
                    Hashtable<Object, Object> hash = new Hashtable<Object, Object>();
                    hash.putAll( (Map<?, ?>) o );
                    returnValue.add( hash );
                }
                else
                {
                    returnValue.add( o );
                }
            }
            result = returnValue;
        }
        else if ( result instanceof Map )
        {
            Hashtable<Object, Object> hash = new Hashtable<Object, Object>();
            hash.putAll( (Map<?, ?>) result );
            result = hash;
        }
        Class<?> returnType = method.getReturnType();

        if ( recording )
        {
            triplets.add( new Triple( method, args, result ) );
        }

        return returnType.cast( result );
    }

    @Override
    public boolean addComment( final String token, final String issueKey, final String comment )
    {
        return proxy.addComment( token, issueKey, comment );
    }

    @Override
    public Hashtable createIssue( final String token, final Hashtable rIssueStruct )
    {
        return proxy.createIssue( token, rIssueStruct );
    }

    @Override
    public Vector getComments( final String token, final String issueKey )
    {
        return proxy.getComments( token, issueKey );
    }

    @Override
    public Vector getComponents( final String token, final String projectKey )
    {
        return proxy.getComponents( token, projectKey );
    }

    @Override
    public Vector getFavouriteFilters( final String token )
    {
        return proxy.getFavouriteFilters( token );
    }

    @Override
    public Hashtable getIssue( final String token, final String issueKey )
    {
        return proxy.getIssue( token, issueKey );
    }

    @Override
    public Vector getIssuesFromFilter( final String token, final String filterId )
    {
        return proxy.getIssuesFromFilter( token, filterId );
    }

    @Override
    public Vector getIssuesFromTextSearch( final String token, final String searchTerms )
    {
        return proxy.getIssuesFromTextSearch( token, searchTerms );
    }

    @Override
    public Vector getIssuesFromTextSearchWithProject( final String token, final Vector projectKeys, final String searchTerms,
                                                      final int maxNumResults )
    {
        return proxy.getIssuesFromTextSearchWithProject( token, projectKeys, searchTerms, maxNumResults );
    }

    @Override
    public Vector getIssueTypes( final String token )
    {
        return proxy.getIssueTypes( token );
    }

    @Override
    public Vector getIssueTypesForProject( final String token, final String projectId )
    {
        return proxy.getIssueTypesForProject( token, projectId );
    }

    @Override
    public Vector getPriorities( final String token )
    {
        return proxy.getPriorities( token );
    }

    @Override
    public Vector getProjectsNoSchemes( final String token )
    {
        return proxy.getProjectsNoSchemes( token );
    }

    @Override
    public Vector getResolutions( final String token )
    {
        return proxy.getResolutions( token );
    }

    @Override
    @Deprecated
    public Vector getSavedFilters( final String token )
    {
        return proxy.getSavedFilters( token );
    }

    @Override
    public Hashtable getServerInfo( final String token )
    {
        return proxy.getServerInfo( token );
    }

    @Override
    public Vector getStatuses( final String token )
    {
        return proxy.getStatuses( token );
    }

    @Override
    public Vector getSubTaskIssueTypes( final String token )
    {
        return proxy.getSubTaskIssueTypes( token );
    }

    @Override
    public Vector getSubTaskIssueTypesForProject( final String token, final String projectId )
    {
        return proxy.getSubTaskIssueTypesForProject( token, projectId );
    }

    @Override
    public Hashtable getUser( final String token, final String username )
    {
        return proxy.getUser( token, username );
    }

    @Override
    public Vector getVersions( final String token, final String projectKey )
    {
        return proxy.getVersions( token, projectKey );
    }

    @Override
    public String login( final String username, final String password )
    {
        return proxy.login( username, password );
    }

    @Override
    public boolean logout( final String token )
    {
        return proxy.logout( token );
    }

    @Override
    public Hashtable updateIssue( final String token, final String issueKey, final Hashtable fieldValues )
    {
        return proxy.updateIssue( token, issueKey, fieldValues );
    }

    private XmlRpcRequest request( final String method, final Object... parameters )
        throws Exception
    {
        return new XmlRpcClientRequestImpl( rpcConfig(), "jira1." + method, parameters );
    }

    private XmlRpcRequestConfig rpcConfig()
        throws Exception
    {
        XmlRpcClientConfigImpl cfg = new XmlRpcClientConfigImpl();
        cfg.setServerURL( url );
        return cfg;
    }

    private XmlRpcClient client()
    {
        if ( this.client == null )
        {
            XmlRpcClientConfigImpl clientConfig = new XmlRpcClientConfigImpl();
            clientConfig.setServerURL( url );
            client = new XmlRpcClient();
            client.setTransportFactory( new XmlRpcAhcTransportFactory( client ) );
            client.setConfig( clientConfig );
        }
        return client;

    }

    @Override
    public void setProxiedUrl( final String url )
    {
        if ( this.url == null || ( !this.url.toString().equalsIgnoreCase( url.toString() ) ) )
        {
            try
            {
                this.url = new URL(url);
            }
            catch ( MalformedURLException e )
            {
                throw new IllegalArgumentException( "Malformed URL: " + url, e );
            }
            this.client = null;
        }
    }

    @Override
    public void write( final File out )
    {
        try
        {
            Recorder.write( triplets, out );
        }
        catch ( IOException e )
        {
            e.printStackTrace();
        }
    }

    @Override
    public void setRecording( final boolean record )
    {
        recording = record;
    }

    @Override
    public boolean isRecording()
    {
        return recording;
    }

}
